from pydantic import BaseModel, ValidationError, validator
import typing
from .discord_types.member import Member

try:
    from typing import Literal
except ImportError:
    # backport
    from typing_extensions import Literal


class IncomingDiscordOption(BaseModel):
    name: str
    value: str


class IncomingDiscordOptionList(BaseModel):
    class Config:
        arbitary_types_allowed = True

    id: str
    name: str
    options: typing.List[IncomingDiscordOption]


class IncomingDiscordInteraction(BaseModel):
    class Config:
        arbitary_types_allowed = True

    type: Literal[2, 3, 4, 5, 6, 7, 8]  # 1 is removed, this lib will handle PING
    id: int
    data: IncomingDiscordOptionList
    guild_id: int
    channel_id: int
    member: Member
    token: str
    version: typing.Optional[Literal[1]] = None



import typing
from fastapi import Response
from loguru import logger
import inspect

from functools import wraps


class EventHandler(object):
    def __init__(self):
        self.callbacks = {}

    def on(self, event, func=None):
        def on(func):
            if not inspect.iscoroutinefunction(func):
                raise TypeError("Function must be a async function.")

            if event not in self.callbacks:
                logger.debug(f"Added {event} to corresponding function to {func}")
                self.callbacks[event] = func
            else:
                raise TypeError("Events can only have one corresponding handler.")

            return func

        return on(func) if func else on

    def check_event_exists(self, event):
        return event in self.callbacks

    async def emit(self, event, *args, **kwargs):
        if event not in self.callbacks:
            raise TypeError(f"event {event} does not have a corresponding handler.")
        return await self.callbacks[event](**kwargs)



from fastapi import FastAPI
from loguru import logger
from .server import router, DiscordVerificationMiddleware
from .server import interaction as router_interaction

import uvicorn

class Dispike(object):
    def __init__(self, client_public_key: str):
        self._internal_application = FastAPI()
        self._internal_application.add_middleware(
            DiscordVerificationMiddleware,
            client_public_key=client_public_key
        )
        self._internal_application.include_router(
            router=router
        )

    @property
    def interaction(self):
        return router_interaction

    @property
    def referenced_application(self):
        return self._internal_application
    
    def run(self, port: int = 5000):
        uvicorn.run(app=self.referenced_application, port=port)


from fastapi import APIRouter, Request, Response
from loguru import logger
from .middlewares.verification import DiscordVerificationMiddleware
from .models.incoming import IncomingDiscordInteraction
from .eventer import EventHandler
import json
import typing

router = APIRouter()
interaction = EventHandler()


@router.post("/interactions")
async def handle_interactions(request: Request) -> Response:
    logger.info("interaction recieved.")

    _get_request_body = json.loads(request.state._cached_body.decode())
    logger.info(_get_request_body)
    if _get_request_body["type"] == 1:
        logger.info("handling ACK Ping.")
        return {"type": 1}

    _parse_to_object = IncomingDiscordInteraction(**_get_request_body)

    if interaction.check_event_exists(_parse_to_object.data.name) == False:
        logger.debug("discarding event not existing.")
        return {}

    if len(_parse_to_object.data.options) > 0:
        arguments = {x.name: x.value for x in _parse_to_object.data.options}
    else:
        arguments = {}

    arguments["ctx"] = _parse_to_object

    return await interaction.emit(_parse_to_object.data.name, **arguments)



from pydantic import BaseModel

try:
    from typing import Literal
except ImportError:
    # backport
    from typing_extensions import Literal


class User(BaseModel):
    class Config:
        arbitary_types_allowed = True

    id: int
    username: str
    avatar: str
    discriminator: str
    public_flags: int



from pydantic import BaseModel
from .helper.embed import Embed
import typing

try:
    from typing import Literal
except ImportError:
    # backport
    from typing_extensions import Literal


class DiscordResponse(object):
    def __init__(
        self,
        content: str = None,
        tts: bool = False,
        embeds: typing.List[Embed] = [],
        show_user_input: bool = False,
    ):

        if content != None:
            if isinstance(content, str) == False:
                raise TypeError(f"Content must be a string")
            elif content == "":
                content = None

        # if isinstance(content, str) == False or content == "" or content != None:
        #    raise TypeError(f"content must be a string. recieved: {content}")

        if isinstance(tts, bool) == False:
            raise TypeError("tts must be a bool")

        self._content = content
        self._tts = tts
        self._embeds = [x.to_dict() for x in embeds]
        if show_user_input == False:
            self._type_response = 3
        else:
            self._type_response = 4

    @property
    def embeds(self):
        return self._embeds

    def add_new_embed(self, embed_to_add: Embed):
        if isinstance(embed_to_add, Embed):
            self._embeds.append(embed_to_add.to_dict())
        else:
            raise TypeError("embed must be a Embed object.")

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, new_content_string: str):
        self._content = new_content_string

    @property
    def tts(self):
        return self._tts

    @tts.setter
    def tts(self, new_tts: bool):
        self._tts = new_tts

    @property
    def response(self):
        if self.content == "":
            self.content = None

        return {
            "type": self._type_response,
            "data": {"tts": self.tts, "content": self.content, "embeds": self.embeds},
        }

    def __call__(self):
        return self.response



from loguru import logger
from fastapi.responses import JSONResponse, Response
from fastapi import HTTPException
from nacl.signing import VerifyKey
from nacl.exceptions import BadSignatureError
from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware

import typing

if typing.TYPE_CHECKING:
    from fastapi import FastAPI


class DiscordVerificationMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: "FastAPI", *, client_public_key: str):
        super().__init__(app)
        self._client_public_key = client_public_key
        logger.info(f"pub: {self._client_public_key}")
        self._verification_key = VerifyKey(bytes.fromhex(self._client_public_key))

    def verify_request(self, passed_signature: str, timestamp: str, body):
        try:
            message = timestamp.encode() + body
            self._verification_key.verify(message, bytes.fromhex(passed_signature))
            return True, 200
        except BadSignatureError:
            logger.error("bad signature")
            return False, 401
        except Exception:
            logger.exception("exception on verifying request")
            return False, 500

    async def dispatch(
        self,
        request: Request,
        call_next: typing.Callable[[Request], typing.Awaitable[Response]],
    ) -> Response:
        logger.debug("intercepting request.")
        try:
            get_signature = request.headers["X-Signature-Ed25519"]
            get_timestamp = request.headers["X-Signature-Timestamp"]

            # https://github.com/encode/starlette/issues/495
            # this is bugged. calling request body will cause a timeout when the next endpoint tries to .body too
            get_body = await request.body()

            # so we need to store the body in the request state.
            request.state._cached_body = get_body
        except Exception:
            # logger.exception("error getting needed data for verification")
            return JSONResponse(
                status_code=400, content={"error_message": "Incorrect request."}
            )

        _status_bool, _status_code = self.verify_request(
            passed_signature=get_signature, timestamp=get_timestamp, body=get_body
        )
        if _status_bool == True:
            logger.info("approved request. forwarding call")
            _dispatch_request = await call_next(request)
            return _dispatch_request

        return JSONResponse(status_code=_status_code)



from pydantic import BaseModel

from datetime import datetime

import typing

try:
    from typing import Literal
except ImportError:
    # backport
    from typing_extensions import Literal

from .user import User


class Member(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    user: User
    roles: typing.List[str]
    premium_since: typing.Union[
        None, str, int
    ]  # ? Discord docs doesn't talk about this so..
    permissions: str
    pending: bool
    nick: typing.Union[None, str]
    mute: bool
    joined_at: datetime
    deaf: bool



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dispike',
 'dispike.helper',
 'dispike.middlewares',
 'dispike.models',
 'dispike.models.discord_types']

package_data = \
{'': ['*']}

install_requires = \
['PyNaCl>=1.4.0,<2.0.0',
 'fastapi>=0.63.0,<0.64.0',
 'httpx>=0.16.1,<0.17.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.7.3,<2.0.0',
 'typing-extensions>=3.7.4,<4.0.0',
 'uvicorn>=0.13.2,<0.14.0']

setup_kwargs = {
    'name': 'dispike',
    'version': '0.2.5a0',
    'description': '',
    'long_description': None,
    'author': 'Mustafa Mohamed',
    'author_email': 'ms7mohamed@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
